@@ -3,17 +3,19 @@
 #include "zeek-config.h"
 
 #include "CompHash.h"
+#include "BroString.h"
+#include "Dict.h"
 #include "Val.h"
+#include "RE.h"
 #include "Reporter.h"
 #include "Func.h"
 
 #include <vector>
 #include <map>
 
-CompositeHash::CompositeHash(TypeList* composite_type)
+CompositeHash::CompositeHash(IntrusivePtr<TypeList> composite_type)
+	: type(std::move(composite_type))
 	{
-	type = composite_type;
-	Ref(type);
 	singleton_tag = TYPE_INTERNAL_ERROR;
 
 	// If the only element is a record, don't treat it as a
@@ -63,7 +65,6 @@ CompositeHash::CompositeHash(TypeList* composite_type)
 
 CompositeHash::~CompositeHash()
 	{
-	Unref(type);
 	delete [] key;
 	}
 
@@ -146,6 +147,30 @@ char* CompositeHash::SingleValHash(int type_check, char* kp0,
 			break;
 			}
 
+		case TYPE_PATTERN:
+			{
+			const char* texts[2] = {
+				v->AsPattern()->PatternText(),
+				v->AsPattern()->AnywherePatternText()
+			};
+
+			uint64_t* kp;
+			for ( int i = 0; i < 2; i++ )
+				{
+				kp = AlignAndPadType<uint64_t>(kp0+i);
+				*kp = strlen(texts[i]) + 1;
+				}
+
+			kp1 = reinterpret_cast<char*>(kp+1);
+			for ( int i = 0; i < 2; i++ )
+				{
+				memcpy(kp1, texts[i], strlen(texts[i]) + 1);
+				kp1 += strlen(texts[i]) + 1;
+				}
+
+			break;
+			}
+
 		case TYPE_RECORD:
 			{
 			char* kp = kp0;
@@ -401,6 +426,19 @@ HashKey* CompositeHash::ComputeSingletonHash(const Val* v, int type_check) const
 		if ( v->Type()->Tag() == TYPE_FUNC )
 			return new HashKey(v->AsFunc()->GetUniqueFuncID());
 
+		if ( v->Type()->Tag() == TYPE_PATTERN )
+			{
+			const char* texts[2] = {
+				v->AsPattern()->PatternText(),
+				v->AsPattern()->AnywherePatternText()
+			};
+			int n = strlen(texts[0]) + strlen(texts[1]) + 2; // 2 for null
+			char* key = new char[n];
+			std::memcpy(key, texts[0], strlen(texts[0]) + 1);
+			std::memcpy(key + strlen(texts[0]) + 1, texts[1], strlen(texts[1]) + 1);
+			return new HashKey(false, key, n);
+			}
+
 		reporter->InternalError("bad index type in CompositeHash::ComputeSingletonHash");
 		return 0;
 
@@ -462,6 +500,17 @@ int CompositeHash::SingleTypeKeySize(BroType* bt, const Val* v,
 			break;
 			}
 
+		case TYPE_PATTERN:
+			{
+			if ( ! v )
+				return (optional && ! calc_static_size) ? sz : 0;
+
+			sz = SizeAlign(sz, 2 * sizeof(uint64_t));
+			sz += strlen(v->AsPattern()->PatternText())
+				+ strlen(v->AsPattern()->AnywherePatternText()) + 2; // 2 for null terminators
+			break;
+			}
+
 		case TYPE_RECORD:
 			{
 			const RecordVal* rv = v ? v->AsRecordVal() : 0;
@@ -831,6 +880,28 @@ const char* CompositeHash::RecoverOneVal(const HashKey* k, const char* kp0,
 			}
 			break;
 
+		case TYPE_PATTERN:
+			{
+			RE_Matcher* re = nullptr;
+			if ( is_singleton )
+				{
+				kp1 = kp0;
+				int divider = strlen(kp0) + 1;
+				re = new RE_Matcher(kp1, kp1 + divider);
+				kp1 += k->Size();
+				}
+			else
+				{
+				const uint64_t* const len = AlignType<uint64_t>(kp0);
+
+				kp1 = reinterpret_cast<const char*>(len+2);
+				re = new RE_Matcher(kp1, kp1 + len[0]);
+				kp1 += len[0] + len[1];
+				}
+			pval = new PatternVal(re);
+			}
+			break;
+
 		case TYPE_RECORD:
 			{
 			const char* kp = kp0;
@@ -848,6 +919,10 @@ const char* CompositeHash::RecoverOneVal(const HashKey* k, const char* kp0,
 
 				kp = RecoverOneVal(k, kp, k_end,
 				                   rt->FieldType(i), v, optional);
+
+				// An earlier call to reporter->InternalError would have called abort() and broken the
+				// call tree that clang-tidy is relying on to get the error described.
+				// NOLINTNEXTLINE(clang-analyzer-core.uninitialized.Branch)
 				if ( ! (v || optional) )
 					{
 					reporter->InternalError("didn't recover expected number of fields from HashKey");
