@@ -3,18 +3,22 @@
 #pragma once
 
 #include <map>
+#include <memory>
 #include <string>
+#include <vector>
 
-#include "Func.h"
-#include "List.h"
-#include "Dict.h"
-#include "EventHandler.h"
+using std::string;
+using std::vector;
+
+class EventHandler;
+class EventHandlerPtr;
+class RE_Matcher;
 
 // The registry keeps track of all events that we provide or handle.
 class EventRegistry {
 public:
-	EventRegistry()		{ }
-	~EventRegistry()	{ }
+	EventRegistry();
+	~EventRegistry() noexcept;
 
 	void Register(EventHandlerPtr handler);
 
@@ -38,7 +42,7 @@ public:
 	void PrintDebug();
 
 private:
-	std::map<std::string, EventHandler*> handlers;
+	std::map<std::string, std::unique_ptr<EventHandler>> handlers;
 };
 
 extern EventRegistry* event_registry;
