@@ -2,17 +2,19 @@
 
 
 #include "zeek-config.h"
+#include "Sessions.h"
 
+#include <netinet/in.h>
 #include <arpa/inet.h>
 
 #include <stdlib.h>
 #include <unistd.h>
 
+#include "Desc.h"
 #include "Net.h"
 #include "Event.h"
 #include "Timer.h"
 #include "NetVar.h"
-#include "Sessions.h"
 #include "Reporter.h"
 
 #include "analyzer/protocol/icmp/ICMP.h"
@@ -28,6 +30,8 @@
 #include "TunnelEncapsulation.h"
 
 #include "analyzer/Manager.h"
+#include "iosource/IOSource.h"
+#include "iosource/PktDumper.h"
 
 // These represent NetBIOS services on ephemeral ports.  They're numbered
 // so that we can use a single int to hold either an actual TCP/UDP server
@@ -39,32 +43,6 @@ enum NetBIOS_Service {
 
 NetSessions* sessions;
 
-void TimerMgrExpireTimer::Dispatch(double t, int is_expire)
-	{
-	if ( mgr->LastAdvance() + timer_mgr_inactivity_timeout < timer_mgr->Time() )
-		{
-		// Expired.
-		DBG_LOG(DBG_TM, "TimeMgr %p has timed out", mgr);
-		mgr->Expire();
-
-		// Make sure events are executed.  They depend on the TimerMgr.
-		::mgr.Drain();
-
-		sessions->timer_mgrs.erase(mgr->GetTag());
-		delete mgr;
-		}
-	else
-		{
-		// Reinstall timer.
-		if ( ! is_expire )
-			{
-			double n = mgr->LastAdvance() +
-					timer_mgr_inactivity_timeout;
-			timer_mgr->Add(new TimerMgrExpireTimer(n, mgr));
-			}
-		}
-	}
-
 void IPTunnelTimer::Dispatch(double t, int is_expire)
 	{
 	NetSessions::IPTunnelMap::const_iterator it =
@@ -142,7 +120,7 @@ void NetSessions::Done()
 
 void NetSessions::NextPacket(double t, const Packet* pkt)
 	{
-	SegmentProfiler(segment_logger, "dispatching-packet");
+	SegmentProfiler prof(segment_logger, "dispatching-packet");
 
 	if ( raw_packet )
 		mgr.QueueEventFast(raw_packet, {pkt->BuildPktHdrVal()});
@@ -207,59 +185,6 @@ void NetSessions::NextPacket(double t, const Packet* pkt)
 		DumpPacket(pkt);
 	}
 
-int NetSessions::CheckConnectionTag(Connection* conn)
-	{
-	if ( current_iosrc->GetCurrentTag() )
-		{
-		// Packet is tagged.
-		if ( conn->GetTimerMgr() == timer_mgr )
-			{
-			// Connection uses global timer queue.  But the
-			// packet has a tag that means we got it externally,
-			// probably from the Time Machine.
-			DBG_LOG(DBG_TM, "got packet with tag %s for already"
-					"known connection, reinstantiating",
-					current_iosrc->GetCurrentTag()->c_str());
-			return 0;
-			}
-		else
-			{
-			// Connection uses local timer queue.
-			TimerMgrMap::iterator i =
-				timer_mgrs.find(*current_iosrc->GetCurrentTag());
-			if ( i != timer_mgrs.end() &&
-			     conn->GetTimerMgr() != i->second )
-				{
-				// Connection uses different local queue
-				// than the tag for the current packet
-				// indicates.
-				//
-				// This can happen due to:
-				//     (1) getting same packets with
-				//		different tags
-				//     (2) timer mgr having already expired
-				DBG_LOG(DBG_TM, "packet ignored due old/inconsistent tag");
-				return -1;
-				}
-
-			return 1;
-			}
-		}
-
-	// Packet is not tagged.
-	if ( conn->GetTimerMgr() != timer_mgr )
-		{
-		// Connection does not use the global timer queue.  That
-		// means that this is a live packet belonging to a
-		// connection for which we have already switched to
-		// processing external input.
-		DBG_LOG(DBG_TM, "packet ignored due to processing it in external data");
-		return -1;
-		}
-
-	return 1;
-	}
-
 static unsigned int gre_header_len(uint16_t flags)
 	{
 	unsigned int len = 4;  // Always has 2 byte flags and 2 byte protocol type.
@@ -732,14 +657,9 @@ void NetSessions::DoNextPacket(double t, const Packet* pkt, const IP_Hdr* ip_hdr
 	else
 		{
 		// We already know that connection.
-		int consistent = CheckConnectionTag(conn);
-		if ( consistent < 0 )
-			return;
-
-		if ( ! consistent || conn->IsReuse(t, data) )
+		if ( conn->IsReuse(t, data) )
 			{
-			if ( consistent )
-				conn->Event(connection_reused, 0);
+			conn->Event(connection_reused, 0);
 
 			Remove(conn);
 			conn = NewConn(key, t, &id, data, proto, ip_hdr->FlowLabel(), pkt, encapsulation);
@@ -1046,9 +966,7 @@ void NetSessions::Remove(Connection* c)
 			}
 
 		c->Done();
-
-		if ( connection_state_remove )
-			c->Event(connection_state_remove, 0);
+		c->RemovalEvent();
 
 		// Zero out c's copy of the key, so that if c has been Ref()'d
 		// up, we know on a future call to Remove() that it's no
@@ -1141,24 +1059,22 @@ void NetSessions::Drain()
 		{
 		Connection* tc = entry.second;
 		tc->Done();
-		tc->Event(connection_state_remove, 0);
+		tc->RemovalEvent();
 		}
 
 	for ( const auto& entry : udp_conns )
 		{
 		Connection* uc = entry.second;
 		uc->Done();
-		uc->Event(connection_state_remove, 0);
+		uc->RemovalEvent();
 		}
 
 	for ( const auto& entry : icmp_conns )
 		{
 		Connection* ic = entry.second;
 		ic->Done();
-		ic->Event(connection_state_remove, 0);
+		ic->RemovalEvent();
 		}
-
-	ExpireTimerMgrs();
 	}
 
 void NetSessions::GetStats(SessionStats& s) const
@@ -1232,25 +1148,9 @@ Connection* NetSessions::NewConn(const ConnIDKey& k, double t, const ConnID* id,
 		return 0;
 		}
 
-	bool external = conn->IsExternal();
-
-	if ( external )
-		conn->AppendAddl(fmt("tag=%s",
-					conn->GetTimerMgr()->GetTag().c_str()));
-
 	if ( new_connection )
-		{
 		conn->Event(new_connection, 0);
 
-		if ( external && connection_external )
-			{
-			conn->ConnectionEventFast(connection_external, 0, {
-				conn->BuildConnVal(),
-				new StringVal(conn->GetTimerMgr()->GetTag().c_str()),
-			});
-			}
-		}
-
 	return conn;
 	}
 
@@ -1337,45 +1237,6 @@ bool NetSessions::WantConnection(uint16_t src_port, uint16_t dst_port,
 	return true;
 	}
 
-TimerMgr* NetSessions::LookupTimerMgr(const TimerMgr::Tag* tag, bool create)
-	{
-	if ( ! tag )
-		{
-		DBG_LOG(DBG_TM, "no tag, using global timer mgr %p", timer_mgr);
-		return timer_mgr;
-		}
-
-	TimerMgrMap::iterator i = timer_mgrs.find(*tag);
-	if ( i != timer_mgrs.end() )
-		{
-		DBG_LOG(DBG_TM, "tag %s, using non-global timer mgr %p", tag->c_str(), i->second);
-		return i->second;
-		}
-	else
-		{
-		if ( ! create )
-			return 0;
-
-		// Create new queue for tag.
-		TimerMgr* mgr = new CQ_TimerMgr(*tag);
-		DBG_LOG(DBG_TM, "tag %s, creating new non-global timer mgr %p", tag->c_str(), mgr);
-		timer_mgrs.insert(TimerMgrMap::value_type(*tag, mgr));
-		double t = timer_mgr->Time() + timer_mgr_inactivity_timeout;
-		timer_mgr->Add(new TimerMgrExpireTimer(t, mgr));
-		return mgr;
-		}
-	}
-
-void NetSessions::ExpireTimerMgrs()
-	{
-	for ( TimerMgrMap::iterator i = timer_mgrs.begin();
-	      i != timer_mgrs.end(); ++i )
-		{
-		i->second->Expire();
-		delete i->second;
-		}
-	}
-
 void NetSessions::DumpPacket(const Packet *pkt, int len)
 	{
 	if ( ! pkt_dumper )
