@@ -2,15 +2,21 @@
 
 #pragma once
 
-#include <list>
-#include <vector>
-
 #include "Tag.h"
 
 #include "../Obj.h"
 #include "../EventHandler.h"
 #include "../Timer.h"
 
+#include <list>
+#include <vector>
+
+#include <sys/types.h> // for u_char
+
+using std::list;
+using std::string;
+
+class BroFile;
 class Rule;
 class Connection;
 class IP_Hdr;
@@ -372,7 +378,7 @@ public:
 	 * @param tag The type of analyzer to add.
 	 * @return the new analyzer instance that was added.
 	 */
-	Analyzer* AddChildAnalyzer(Tag tag);
+	Analyzer* AddChildAnalyzer(const Tag& tag);
 
 	/**
 	 * Removes a child analyzer. It's ok for the analyzer to not to be a
@@ -641,7 +647,7 @@ protected:
 	 *
 	 * @param orig True if asking about the originator side.
 	 */
-	bool HasSupportAnalyzer(Tag tag, bool orig);
+	bool HasSupportAnalyzer(const Tag& tag, bool orig);
 
 	/**
 	 * Returns the first still active support analyzer for the given
