@@ -1,14 +1,20 @@
 // See the file "COPYING" in the main distribution directory for copyright.
 
-#include <algorithm>
+#include "analyzer/protocol/tcp/TCP.h"
 
+#include <vector>
+
+#include "analyzer/protocol/tcp/TCP_Reassembler.h"
+#include "analyzer/protocol/pia/PIA.h"
+
+#include "IP.h"
+#include "Net.h"
 #include "NetVar.h"
 #include "File.h"
 #include "Event.h"
-
-#include "analyzer/protocol/pia/PIA.h"
-#include "analyzer/protocol/tcp/TCP.h"
-#include "analyzer/protocol/tcp/TCP_Reassembler.h"
+#include "Reporter.h"
+#include "Sessions.h"
+#include "DebugLogger.h"
 
 #include "events.bif.h"
 #include "types.bif.h"
@@ -1190,6 +1196,19 @@ void TCP_Analyzer::DeliverPacket(int len, const u_char* data, bool is_orig,
 	if ( (tcp_option || tcp_options) && tcp_hdr_len > sizeof(*tp) )
 		ParseTCPOptions(tp, is_orig);
 
+	// PIA/signature matching state needs to be initialized before
+	// processing/reassembling any TCP data, since that processing may
+	// itself try to perform signature matching.  Also note that a SYN
+	// packet may technically carry data (see RFC793 Section 3.4 and also
+	// TCP Fast Open).
+	CheckPIA_FirstPacket(is_orig, ip);
+
+	// Note the similar/inverse logic to connection_attempt.
+	if ( resp->state != TCP_ENDPOINT_INACTIVE ||
+	     (orig->state != TCP_ENDPOINT_SYN_SENT &&
+	      orig->state != TCP_ENDPOINT_SYN_ACK_SENT))
+		Conn()->SetSuccessful();
+
 	if ( DEBUG_tcp_data_sent )
 		{
 		DEBUG_MSG("%.6f before DataSent: len=%d caplen=%d skip=%d\n",
@@ -1243,8 +1262,6 @@ void TCP_Analyzer::DeliverPacket(int len, const u_char* data, bool is_orig,
 
 	if ( ! reassembling )
 		ForwardPacket(len, data, is_orig, rel_data_seq, ip, caplen);
-
-	CheckPIA_FirstPacket(is_orig, ip);
 	}
 
 void TCP_Analyzer::DeliverStream(int len, const u_char* data, bool orig)
