@@ -1,3 +1,4 @@
+#include "Manager.h"
 
 #include <broker/broker.hh>
 #include <broker/zeek.hh>
@@ -5,12 +6,13 @@
 #include <cstring>
 #include <unistd.h>
 
-#include "Manager.h"
 #include "Data.h"
 #include "Store.h"
 #include "util.h"
 #include "Var.h"
+#include "Desc.h"
 #include "Reporter.h"
+#include "IntrusivePtr.h"
 #include "broker/comm.bif.h"
 #include "broker/data.bif.h"
 #include "broker/messaging.bif.h"
@@ -19,6 +21,7 @@
 #include "DebugLogger.h"
 #include "iosource/Manager.h"
 #include "SerializationFormat.h"
+#include "Net.h"
 
 using namespace std;
 
@@ -65,12 +68,6 @@ const broker::endpoint_info Manager::NoPeer{{}, {}};
 
 int Manager::script_scope = 0;
 
-struct unref_guard {
-	unref_guard(Val* v) : val(v) {}
-	~unref_guard() { Unref(val); }
-	Val* val;
-};
-
 struct scoped_reporter_location {
 	scoped_reporter_location(Frame* frame)
 		{
@@ -133,15 +130,11 @@ Manager::Manager(bool arg_reading_pcaps)
 	reading_pcaps = arg_reading_pcaps;
 	after_zeek_init = false;
 	peer_count = 0;
-	times_processed_without_idle = 0;
 	log_batch_size = 0;
-	log_batch_interval = 0;
 	log_topic_func = nullptr;
 	vector_of_data_type = nullptr;
 	log_id_type = nullptr;
 	writer_id_type = nullptr;
-
-	SetIdle(false);
 	}
 
 Manager::~Manager()
@@ -153,7 +146,6 @@ void Manager::InitPostScript()
 	DBG_LOG(DBG_BROKER, "Initializing");
 
 	log_batch_size = get_option("Broker::log_batch_size")->AsCount();
-	log_batch_interval = get_option("Broker::log_batch_interval")->AsInterval();
 	default_log_topic_prefix =
 	    get_option("Broker::default_log_topic_prefix")->AsString()->CheckString();
 	log_topic_func = get_option("Broker::log_topic")->AsFunc();
@@ -215,12 +207,20 @@ void Manager::InitPostScript()
 
 	auto cqs = get_option("Broker::congestion_queue_size")->AsCount();
 	bstate = std::make_shared<BrokerState>(std::move(config), cqs);
+
+	if ( ! iosource_mgr->RegisterFd(bstate->subscriber.fd(), this) )
+		reporter->FatalError("Failed to register broker subscriber with iosource_mgr");
+	if ( ! iosource_mgr->RegisterFd(bstate->status_subscriber.fd(), this) )
+		reporter->FatalError("Failed to register broker status subscriber with iosource_mgr");
 	}
 
 void Manager::Terminate()
 	{
 	FlushLogBuffers();
 
+	iosource_mgr->UnregisterFd(bstate->subscriber.fd(), this);
+	iosource_mgr->UnregisterFd(bstate->status_subscriber.fd(), this);
+
 	vector<string> stores_to_close;
 
 	for ( auto& x : data_stores )
@@ -281,8 +281,6 @@ void Manager::FlushPendingQueries()
 				}
 			}
 		}
-
-	SetIdle(false);
 	}
 
 uint16_t Manager::Listen(const string& addr, uint16_t port)
@@ -396,7 +394,7 @@ bool Manager::PublishEvent(string topic, RecordVal* args)
 		xs.emplace_back(data_val->data);
 		}
 
-	return PublishEvent(topic, event_name, std::move(xs));
+	return PublishEvent(std::move(topic), event_name, std::move(xs));
 	}
 
 bool Manager::PublishIdentifier(std::string topic, std::string id)
@@ -580,8 +578,7 @@ bool Manager::PublishLogWrite(EnumVal* stream, EnumVal* writer, string path, int
 	auto& pending_batch = lb.msgs[topic];
 	pending_batch.emplace_back(msg.move_data());
 
-	if ( lb.message_count >= log_batch_size ||
-	     (network_time - lb.last_flush >= log_batch_interval ) )
+	if ( lb.message_count >= log_batch_size )
 		statistics.num_logs_outgoing += lb.Flush(bstate->endpoint, log_batch_size);
 
 	return true;
@@ -608,7 +605,6 @@ size_t Manager::LogBuffer::Flush(broker::endpoint& endpoint, size_t log_batch_si
 		}
 
 	auto rval = message_count;
-	last_flush = network_time;
 	message_count = 0;
 	return rval;
 	}
@@ -619,8 +615,9 @@ size_t Manager::FlushLogBuffers()
 	auto rval = 0u;
 
 	for ( auto& lb : log_buffers )
-		rval += lb.Flush(bstate->endpoint, log_batch_interval);
+		rval += lb.Flush(bstate->endpoint, log_batch_size);
 
+	statistics.num_logs_outgoing += rval;
 	return rval;
 	}
 
@@ -628,7 +625,7 @@ void Manager::Error(const char* format, ...)
 	{
 	va_list args;
 	va_start(args, format);
-	auto msg = fmt(format, args);
+	auto msg = vfmt(format, args);
 	va_end(args);
 
 	if ( script_scope )
@@ -821,24 +818,6 @@ bool Manager::Unsubscribe(const string& topic_prefix)
 	return true;
 	}
 
-void Manager::GetFds(iosource::FD_Set* read, iosource::FD_Set* write,
-                           iosource::FD_Set* except)
-	{
-	read->Insert(bstate->subscriber.fd());
-	read->Insert(bstate->status_subscriber.fd());
-
-	for ( auto& x : data_stores )
-		read->Insert(x.second->proxy.mailbox().descriptor());
-	}
-
-double Manager::NextTimestamp(double* local_network_time)
-	{
-	// We're only asked for a timestamp if either (1) a FD was ready
-	// or (2) we're not idle (and we go idle if when Process is no-op),
-	// so there's no case where returning -1 to signify a skip will help.
-	return timer_mgr->Time();
-	}
-
 void Manager::DispatchMessage(const broker::topic& topic, broker::data msg)
 	{
 	switch ( broker::zeek::Message::type(msg) ) {
@@ -891,6 +870,10 @@ void Manager::DispatchMessage(const broker::topic& topic, broker::data msg)
 
 void Manager::Process()
 	{
+	// Ensure that time gets update before processing broker messages, or events
+	// based on them might get scheduled wrong.
+	net_update_time(current_time());
+
 	bool had_input = false;
 
 	auto status_msgs = bstate->status_subscriber.poll();
@@ -955,30 +938,6 @@ void Manager::Process()
 			// network_time, may as well do so now because otherwise the
 			// broker/cluster logs will end up using timestamp 0.
 			net_update_time(current_time());
-
-		++times_processed_without_idle;
-
-		// The max number of Process calls allowed to happen in a row without
-		// idling is chosen a bit arbitrarily, except 12 is around half of the
-		// SELECT_FREQUENCY (25).
-		//
-		// But probably the general idea should be for it to have some relation
-		// to the SELECT_FREQUENCY: less than it so other busy IOSources can
-		// fit several Process loops in before the next poll event (e.g. the
-		// select() call ), but still large enough such that we don't have to
-		// wait long before the next poll ourselves after being forced to idle.
-		if ( times_processed_without_idle > 12 )
-			{
-			times_processed_without_idle = 0;
-			SetIdle(true);
-			}
-		else
-			SetIdle(false);
-		}
-	else
-		{
-		times_processed_without_idle = 0;
-		SetIdle(true);
 		}
 	}
 
@@ -1039,7 +998,7 @@ void Manager::ProcessEvent(const broker::topic& topic, broker::zeek::Event ev)
 		auto val = data_to_val(std::move(args[i]), expected_type);
 
 		if ( val )
-			vl.push_back(val);
+			vl.push_back(val.release());
 		else
 			{
 			auto expected_name = type_name(expected_type->Tag());
@@ -1086,8 +1045,6 @@ bool bro_broker::Manager::ProcessLogCreate(broker::zeek::LogCreate lc)
 		return false;
 		}
 
-	unref_guard stream_id_unreffer{stream_id};
-
 	auto writer_id = data_to_val(std::move(lc.writer_id()), writer_id_type);
 	if ( ! writer_id )
 		{
@@ -1095,8 +1052,6 @@ bool bro_broker::Manager::ProcessLogCreate(broker::zeek::LogCreate lc)
 		return false;
 		}
 
-	unref_guard writer_id_unreffer{writer_id};
-
 	auto writer_info = std::unique_ptr<logging::WriterBackend::WriterInfo>(new logging::WriterBackend::WriterInfo);
 	if ( ! writer_info->FromBroker(std::move(lc.writer_info())) )
 		{
@@ -1163,8 +1118,6 @@ bool bro_broker::Manager::ProcessLogWrite(broker::zeek::LogWrite lw)
 		return false;
 		}
 
-	unref_guard stream_id_unreffer{stream_id};
-
 	// Get writer ID.
 	auto writer_id = data_to_val(std::move(lw.writer_id()), writer_id_type);
 	if ( ! writer_id )
@@ -1173,7 +1126,6 @@ bool bro_broker::Manager::ProcessLogWrite(broker::zeek::LogWrite lw)
 		return false;
 		}
 
-	unref_guard writer_id_unreffer{writer_id};
 	auto path = caf::get_if<std::string>(&lw.path());
 
 	if ( ! path )
@@ -1258,7 +1210,7 @@ bool Manager::ProcessIdentifierUpdate(broker::zeek::IdentifierUpdate iu)
 		return false;
 		}
 
-	id->SetVal(val);
+	id->SetVal(std::move(val));
 	return true;
 	}
 
@@ -1497,6 +1449,7 @@ StoreHandleVal* Manager::MakeMaster(const string& name, broker::backend type,
 	Ref(handle);
 
 	data_stores.emplace(name, handle);
+	iosource_mgr->RegisterFd(handle->proxy.mailbox().descriptor(), this);
 
 	if ( bstate->endpoint.use_real_time() )
 		return handle;
@@ -1533,6 +1486,7 @@ StoreHandleVal* Manager::MakeClone(const string& name, double resync_interval,
 	Ref(handle);
 
 	data_stores.emplace(name, handle);
+	iosource_mgr->RegisterFd(handle->proxy.mailbox().descriptor(), this);
 
 	return handle;
 	}
@@ -1551,6 +1505,8 @@ bool Manager::CloseStore(const string& name)
 	if ( s == data_stores.end() )
 		return false;
 
+	iosource_mgr->UnregisterFd(s->second->proxy.mailbox().descriptor(), this);
+
 	for ( auto i = pending_queries.begin(); i != pending_queries.end(); )
 		if ( i->second->Store().name() == name )
 			{
