@@ -1,17 +1,16 @@
 // See the file  in the main distribution directory for copyright.
 
-
-#include "plugin/Plugin.h"
-
 #include "X509.h"
 #include "OCSP.h"
+#include "plugin/Plugin.h"
+#include "file_analysis/Component.h"
 
 namespace plugin {
 namespace Zeek_X509 {
 
 class Plugin : public plugin::Plugin {
 public:
-	plugin::Configuration Configure()
+	plugin::Configuration Configure() override
 		{
 		AddComponent(new ::file_analysis::Component("X509", ::file_analysis::X509::Instantiate));
 		AddComponent(new ::file_analysis::Component("OCSP_REQUEST", ::file_analysis::OCSP::InstantiateRequest));
@@ -22,6 +21,12 @@ public:
 		config.description = "X509 and OCSP analyzer";
 		return config;
 		}
+
+	void Done() override
+		{
+		plugin::Plugin::Done();
+		::file_analysis::X509::FreeRootStore();
+		}
 } plugin;
 
 }
