@@ -1,12 +1,17 @@
 // See the file "COPYING" in the main distribution directory for copyright.
 
-#include <algorithm>
-
 #include "Manager.h"
+
+#include <string>
+#include <utility>
+
 #include "ReaderFrontend.h"
 #include "ReaderBackend.h"
+#include "Desc.h"
+#include "module_util.h"
 #include "input.bif.h"
 
+#include "Expr.h"
 #include "Event.h"
 #include "EventHandler.h"
 #include "NetVar.h"
@@ -444,6 +449,7 @@ bool Manager::CreateEventStream(RecordVal* fval)
 	if ( status )
 		{
 		reporter->Error("Input stream %s: Problem unrolling", stream_name.c_str());
+		for ( auto& f : fieldsV ) delete f;
 		return false;
 		}
 
@@ -453,6 +459,7 @@ bool Manager::CreateEventStream(RecordVal* fval)
 	if ( ! res )
 		{
 		delete stream;
+		for ( auto& f : fieldsV ) delete f;
 		return false;
 		}
 
@@ -671,6 +678,7 @@ bool Manager::CreateTableStream(RecordVal* fval)
 	if ( (valfields > 1) && (want_record->InternalInt() != 1) )
 		{
 		reporter->Error("Input stream %s: Stream does not want a record (want_record=F), but has more then one value field.", stream_name.c_str());
+		for ( auto& f : fieldsV ) delete f;
 		return false;
 		}
 
@@ -680,6 +688,7 @@ bool Manager::CreateTableStream(RecordVal* fval)
 	if ( status )
 		{
 		reporter->Error("Input stream %s: Problem unrolling", stream_name.c_str());
+		for ( auto& f : fieldsV ) delete f;
 		return false;
 		}
 
@@ -689,6 +698,7 @@ bool Manager::CreateTableStream(RecordVal* fval)
 		if ( ! res )
 			{
 			delete stream;
+			for ( auto& f : fieldsV ) delete f;
 			return false;
 			}
 		}
@@ -1289,7 +1299,8 @@ int Manager::SendEntryTable(Stream* i, const Value* const *vals)
 	if ( predidx != 0 )
 		Unref(predidx);
 
-	stream->currDict->Insert(idxhash, ih);
+	auto prev = stream->currDict->Insert(idxhash, ih);
+	delete prev;
 	delete idxhash;
 
 	if ( stream->event )
